from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class PasswdData (BoundData):
    pass


class UserData (BoundData):
    fields = ['name', 'password', 'uid', 'gid', 'comment', 'home', 'shell']

    def template(self):
        return Node(
            'line',
            *[
                Node('token', children=[
                    PropertyNode('value', '')
                ])
                for x in UserData.fields
            ]
        )


PasswdData.bind_collection('users', item_class=UserData)
for i in range(0, len(UserData.fields)):
    path = lambda i: lambda x: x.children[i]
    UserData.bind_property('value', UserData.fields[i], path=path(i))



yn_getter = lambda x: x == 'yes'
yn_setter = lambda x: 'yes' if x else 'no'
tf_getter = lambda x: x == 'true'
tf_setter = lambda x: 'true' if x else 'false'
onezero_getter = lambda x: x == '1'
onezero_setter = lambda x: '1' if x else '0'



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class DHCPDData (BoundData):
    pass


class SubnetData (BoundData):
    def template(self):
        return Node(
            'subnet',
            parameter='192.168.0.0 netmask 255.255.255.0',
        )


class RangeData (BoundData):
    def template(self):
        return PropertyNode('range', '192.168.0.1 192.168.0.100')


class OptionData (BoundData):
    def template(self):
        return PropertyNode('option', '')


DHCPDData.bind_collection('subnets', selector=lambda x: x.name == 'subnet', item_class=SubnetData)
SubnetData.bind_attribute('parameter', 'name')
SubnetData.bind_collection('subnets', selector=lambda x: x.name == 'subnet', item_class=SubnetData)
SubnetData.bind_collection('ranges', selector=lambda x: x.name == 'range', item_class=RangeData)
RangeData.bind_attribute('value', 'range')
OptionData.bind_attribute('value', 'value')

for x in [DHCPDData, SubnetData]:
    x.bind_collection('options', selector=lambda x: x.name == 'option', item_class=OptionData)



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class HostsData (BoundData):
    pass


class HostData (BoundData):
    def template(self):
        return Node('line', children=[
            Node('token', children=[PropertyNode('value', '127.0.0.1')]),
            Node('token', children=[PropertyNode('value', 'localhost')]),
        ])


class AliasData (BoundData):
    def template(self):
        return Node()


HostsData.bind_collection('hosts', item_class=HostData)
HostData.bind_property('value', 'address', path=lambda x: x.children[0])
HostData.bind_property('value', 'name', path=lambda x: x.children[1])
HostData.bind_collection('aliases', item_class=AliasData, selector=lambda x: x.parent.indexof(x) > 1)
AliasData.bind_property('value', 'name')



from reconfigure.nodes import *
from reconfigure.parsers import BaseParser


class NSDParser (BaseParser):
    """
    A parser for NSD DNS server nsd.conf file
    """

    def parse(self, content):
        lines = content.splitlines()
        root = RootNode()
        last_comment = None
        node = root
        for line in lines:
            line = line.strip()
            if line:
                if line.startswith('#'):
                    c = line.strip('#').strip()
                    if last_comment:
                        last_comment += '\n' + c
                    else:
                        last_comment = c
                    continue

                key, value = line.split(':')
                value = value.strip()
                key = key.strip()
                if key in ['server', 'zone', 'key']:
                    node = Node(key, comment=last_comment)
                    root.append(node)
                else:
                    node.append(PropertyNode(key, value, comment=last_comment))
                last_comment = None
        return root

    def stringify_comment(self, line, comment):
        if comment:
            return ''.join('# %s\n' % x for x in comment.splitlines()) + line
        return line

    def stringify(self, tree):
        r = ''
        for node in tree.children:
            r += self.stringify_comment(node.name + ':', node.comment) + '\n'
            for subnode in node.children:
                l = '%s: %s' % (subnode.name, subnode.value)
                r += self.stringify_comment(l, subnode.comment) + '\n'
            r += '\n'
        return r



class Node (object):
    """
    A base node class for the Node Tree.
    This class represents a named container node.
    """

    def __init__(self, name=None, *args, **kwargs):
        """
        :param name: Node name
        :param *args: Children
        :param comment: Node comment string
        :param origin: Node's source location (usually path to the file)
        """
        self.name = name
        self.origin = None
        self.children = []
        self._extra_content = kwargs.pop('extra_content', None)
        for node in list(args) + kwargs.pop('children', []):
            self.append(node)
        self.comment = kwargs.pop('comment', None)
        self.__dict__.update(kwargs)

    def __str__(self):
        s = '(%s)' % self.name
        if self.comment:
            s += ' (%s)' % self.comment
        s += '\n'
        for child in self.children:
            if child.origin != self.origin:
                s += '\t@%s\n' % child.origin
            s += '\n'.join('\t' + x for x in str(child).splitlines()) + '\n'
        return s

    def __hash__(self):
        return sum(hash(x) for x in [self.name, self.origin, self.comment] + self.children)

    def __eq__(self, other):
        if other is None:
            return False

        return \
            self.name == other.name and \
            self.comment == other.comment and \
            self.origin == other.origin and \
            set(self.children) == set(other.children)

    def __iter__(self):
        return iter(self.children)

    def __len__(self):
        return len(self.children)

    def __nonzero__(self):
        return True

    def __getitem__(self, key):
        if type(key) in (int, slice):
            return self.children[key]
        return self.get(key)

    def __setitem__(self, key, value):
        if type(key) is int:
            self.children[key] = value
        self.set_property(key, value)

    def __contains__(self, item):
        return item in self.children

    def indexof(self, node):
        """
        :returns: index of the node in the children array or ``None`` if it's not a child
        """
        if node in self.children:
            return self.children.index(node)
        else:
            return None

    def get(self, name, default=None):
        """
        :returns: a child node by its name or ``default``
        """
        for child in self.children:
            if child.name == name:
                return child
        if default:
            self.append(default)
        return default

    def get_all(self, name):
        """
        :returns: list of child nodes with supplied ``name``
        """
        return [n for n in self.children if n.name == name]

    def append(self, node):
        if not node.origin:
            node.origin = self.origin
        self.children.append(node)
        node.parent = self

    def remove(self, node):
        self.children.remove(node)

    def replace(self, name, node=None):
        """
        Replaces the child nodes by ``name``

        :param node: replacement node or list of nodes

        ::

            n.append(Node('a'))
            n.append(Node('a'))
            n.replace('a', None)
            assert(len(n.get_all('a')) == 0)

        """
        if name:
            self.children = [c for c in self.children if c.name != name]
        if node is not None:
            if type(node) == list:
                for n in node:
                    self.children.append(n)
            else:
                self.children.append(node)

    def set_property(self, name, value):
        """
        Creates or replaces a child :class:`PropertyNode` by name.
        """
        node = self.get(name)
        if node is None:
            node = PropertyNode(name, value)
            self.append(node)
        node.value = value
        return self


class RootNode (Node):
    """
    A special node class that indicates tree root
    """


class PropertyNode (Node):
    """
    A node that serves as a property of its parent node.
    """

    def __init__(self, name, value, comment=None):
        """
        :param name: Property name
        :param value: Property value
        """
        Node.__init__(self, name, comment=comment)
        self.value = value

    def __eq__(self, other):
        if other is None:
            return False

        return \
            Node.__eq__(self, other) and \
            self.value == other.value

    def __hash__(self):
        return Node.__hash__(self) + hash(self.value)

    def __str__(self):
        s = '%s = %s' % (self.name, self.value)
        if self.comment:
            s += ' (%s)' % self.comment
        return s


class IncludeNode (Node):
    """
    A node that indicates a junction point between two config files
    """

    def __init__(self, files):
        """
        :param files: an includer-dependent config location specifier
        """
        Node.__init__(self)
        self.name = '<include>'
        self.files = files

    def __str__(self):
        return '<include> %s' % self.files



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class ResolvData (BoundData):
    pass


class ItemData (BoundData):
    def template(self):
        return Node('line', children=[
            Node('token', children=[PropertyNode('value', 'nameserver')]),
            Node('token', children=[PropertyNode('value', '8.8.8.8')]),
        ])


ResolvData.bind_collection('items', item_class=ItemData)
ItemData.bind_property('value', 'name', path=lambda x: x.children[0])
ItemData.bind_property('value', 'value', path=lambda x: x.children[1])



from reconfigure.includers.base import BaseIncluder
from reconfigure.nodes import *
import glob
import os


class AutoIncluder (BaseIncluder):
    """
    This base includer automatically walks the node tree and loads the include files from ``IncludeNode.files`` properties. ``files`` is supposed to contain absolute path, relative path or a shell wildcard.
    """

    def compose(self, origin, tree):
        self.compose_rec(origin, origin, tree)
        return tree

    def compose_rec(self, root, origin, node):
        if not node.origin:
            node.origin = origin
        for child in node.children:
            self.compose_rec(root, origin, child)
        for child in node.children:
            spec = self.is_include(child)
            if spec:
                files = spec
                if node.origin and not files.startswith('/'):
                    files = os.path.join(os.path.split(root)[0], files)
                if '*' in files or '.' in files:
                    files = glob.glob(files)
                else:
                    files = [files]
                for file in files:
                    if file in self.content_map:
                        content = self.content_map[file]
                    else:
                        content = open(file, 'r').read()
                    subtree = self.parser.parse(content)
                    node.children.extend(subtree.children)
                    self.compose_rec(root, file, subtree)
                node.children[node.children.index(child)] = IncludeNode(spec)

    def decompose(self, tree):
        result = {}
        result[tree.origin] = self.decompose_rec(tree, result)
        return result

    def decompose_rec(self, node, result):
        for child in node.children:
            if child.__class__ == IncludeNode:
                replacement = self.remove_include(child)
                if replacement:
                    node.children[node.children.index(child)] = replacement
            else:
                if child.origin is None:
                    child.origin = node.origin
                if child.origin != node.origin:
                    node.children.remove(child)
                    result.setdefault(child.origin, RootNode()).children.append(self.decompose_rec(child, result))
                else:
                    self.decompose_rec(child, result)
        return node

    def is_include(self, node):
        """
        Should return whether the node is an include node and return file pattern glob if it is
        """

    def remove_include(self, node):
        """
        Shoud transform :class:`reconfigure.nodes.IncludeNode` into a normal Node to be stringified into the file
        """



import json


class BoundCollection (object):
    """
    Binds a list-like object to a set of nodes

    :param node: target node (its children will be bound)
    :param item_class: :class:`BoundData` class for items
    :param selector: ``lambda x: bool``, used to filter out a subset of nodes
    """

    def __init__(self, node, item_class, selector=lambda x: True):
        self.node = node
        self.selector = selector
        self.item_class = item_class
        self.data = []
        self.rebuild()

    def rebuild(self):
        """
        Discards cached collection and rebuilds it from the nodes
        """
        del self.data[:]
        for node in self.node.children:
            if self.selector(node):
                self.data.append(self.item_class(node))

    def to_dict(self):
        return [x.to_dict() if hasattr(x, 'to_dict') else x for x in self]

    def to_json(self):
        return json.dumps(self.to_dict(), indent=4)

    def __str__(self):
        return self.to_json()

    def __iter__(self):
        return self.data.__iter__()

    def __getitem__(self, index):
        return self.data.__getitem__(index)

    def __len__(self):
        return len(self.data)

    def __contains__(self, item):
        return item in self.data

    def append(self, item):
        self.node.append(item._node)
        self.data.append(item)

    def remove(self, item):
        self.node.remove(item._node)
        self.data.remove(item)

    def insert(self, index, item):
        self.node.children.insert(index, item._node)
        self.data.insert(index, item)

    def pop(self, index):
        d = self[index]
        self.remove(d)
        return d


class BoundDictionary (BoundCollection):
    """
    Binds a dict-like object to a set of nodes. Accepts same params as :class:`BoundCollection` plus ``key``

    :param key: ``lambda value: object``, is used to get key for value in the collection
    """

    def __init__(self, key=None, **kwargs):
        self.key = key
        BoundCollection.__init__(self, **kwargs)

    def rebuild(self):
        BoundCollection.rebuild(self)
        self.rebuild_dict()

    def rebuild_dict(self):
        self.datadict = dict((self.key(x), x) for x in self.data)

    def to_dict(self):
        return dict((k, x.to_dict() if hasattr(x, 'to_dict') else x) for k, x in self.items())

    def __getitem__(self, key):
        self.rebuild_dict()
        return self.datadict[key]

    def __setitem__(self, key, value):
        self.rebuild_dict()
        if not key in self:
            self.append(value)
        self.datadict[key] = value
        self.rebuild_dict()

    def __contains__(self, key):
        self.rebuild_dict()
        return key in self.datadict

    def __iter__(self):
        self.rebuild_dict()
        return self.datadict.__iter__()

    def iteritems(self):
        self.rebuild_dict()
        return self.datadict.items()

    items = iteritems

    def setdefault(self, k, v):
        if not k in self:
            self[k] = v
            self.append(v)
        return self[k]

    def values(self):
        self.rebuild_dict()
        return self.data

    def update(self, other):
        for k, v in other.items():
            self[k] = v

    def pop(self, key):
        if key in self:
            self.remove(self[key])


class BoundData (object):
    """
    Binds itself to a node.

    ``bind_*`` classmethods should be called on module-level, after subclass declaration.

    :param node: all bindings will be relative to this node
    :param kwargs: if ``node`` is ``None``, ``template(**kwargs)`` will be used to create node tree fragment
    """

    def __init__(self, node=None, **kwargs):
        if node is None:
            node = self.template(**kwargs)
        self._node = node
        self.bind_attribute('_extra_content', '_extra_content')

    def template(self, **kwargs):
        """
        Override to create empty objects.

        :returns: a :class:`reconfigure.nodes.Node` tree that will be used as a template for new BoundData instance
        """
        return None

    def to_dict(self):
        res_dict = {}
        for attr_key in self.__class__.__dict__:
            if attr_key in self.__class__._bound:
                if attr_key == '_extra_content':
                    continue
                attr_value = getattr(self, attr_key)
                if isinstance(attr_value, BoundData):
                    res_dict[attr_key] = attr_value.to_dict()
                elif isinstance(attr_value, BoundCollection):
                    res_dict[attr_key] = attr_value.to_dict()
                else:
                    res_dict[attr_key] = attr_value
        return res_dict

    def to_json(self):
        return json.dumps(self.to_dict(), indent=4)

    def __str__(self):
        return self.to_json()

    @classmethod
    def bind(cls, data_property, getter, setter):
        """
        Creates an arbitrary named property in the class with given getter and setter. Not usually used directly.

        :param data_property: property name
        :param getter: ``lambda: object``, property getter
        :param setter: ``lambda value: None``, property setter
        """
        if not hasattr(cls, '_bound'):
            cls._bound = []
        cls._bound.append(data_property)
        setattr(cls, data_property, property(getter, setter))

    @classmethod
    def bind_property(cls, node_property, data_property, default=None, \
            default_remove=[], \
            path=lambda x: x, getter=lambda x: x, setter=lambda x: x):
        """
        Binds the value of a child :class:`reconfigure.node.PropertyNode` to a property

        :param node_property: ``PropertyNode``'s ``name``
        :param data_property: property name to be created
        :param default: default value of the property (is ``PropertyNode`` doesn't exist)
        :param default_remove: if setting a value contained in default_remove, the target property is removed
        :param path: ``lambda self.node: PropertyNode``, can be used to point binding to another Node instead of ``self.node``.
        :param getter: ``lambda object: object``, used to transform value when getting
        :param setter: ``lambda object: object``, used to transform value when setting
        """
        def pget(self):
            prop = path(self._node).get(node_property)
            if prop is not None:
                return getter(prop.value)
            else:
                return default

        def pset(self, value):
            if setter(value) in default_remove:
                node = path(self._node).get(node_property)
                if node is not None:
                    path(self._node).remove(node)
            else:
                path(self._node).set_property(node_property, setter(value))

        cls.bind(data_property, pget, pset)

    @classmethod
    def bind_attribute(cls, node_attribute, data_property, default=None, \
            path=lambda x: x, getter=lambda x: x, setter=lambda x: x):
        """
        Binds the value of node object's attribute to a property

        :param node_attribute: ``Node``'s attribute name
        :param data_property: property name to be created
        :param default: default value of the property (is ``PropertyNode`` doesn't exist)
        :param path: ``lambda self.node: PropertyNode``, can be used to point binding to another Node instead of ``self.node``.
        :param getter: ``lambda object: object``, used to transform value when getting
        :param setter: ``lambda object: object``, used to transform value when setting
        """
        def pget(self):
            prop = getattr(path(self._node), node_attribute)
            if prop is not None:
                return getter(prop)
            else:
                return getter(default)

        def pset(self, value):
            setattr(path(self._node), node_attribute, setter(value))

        cls.bind(data_property, pget, pset)

    @classmethod
    def bind_collection(cls, data_property, path=lambda x: x, selector=lambda x: True, item_class=None, \
        collection_class=BoundCollection, **kwargs):
        """
        Binds the subset of node's children to a collection property

        :param data_property: property name to be created
        :param path: ``lambda self.node: PropertyNode``, can be used to point binding to another Node instead of ``self.node``.
        :param selector: ``lambda Node: bool``, can be used to filter out a subset of child nodes
        :param item_class: a :class:`BoundData` subclass to be used for collection items
        :param collection_class: a :class:`BoundCollection` subclass to be used for collection property itself
        """
        def pget(self):
            if not hasattr(self, '__' + data_property):
                setattr(self, '__' + data_property,
                    collection_class(
                        node=path(self._node),
                        item_class=item_class,
                        selector=selector,
                        **kwargs
                    )
                )
            return getattr(self, '__' + data_property)

        cls.bind(data_property, pget, None)

    @classmethod
    def bind_name(cls, data_property, getter=lambda x: x, setter=lambda x: x):
        """
        Binds the value of node's ``name`` attribute to a property

        :param data_property: property name to be created
        :param getter: ``lambda object: object``, used to transform value when getting
        :param setter: ``lambda object: object``, used to transform value when setting
        """
        def pget(self):
            return getter(self._node.name)

        def pset(self, value):
            self._node.name = setter(value)

        cls.bind(data_property, pget, pset)

    @classmethod
    def bind_child(cls, data_property, path=lambda x: x, item_class=None):
        """
        Directly binds a child Node to a BoundData property

        :param data_property: property name to be created
        :param path: ``lambda self.node: PropertyNode``, can be used to point binding to another Node instead of ``self.node``.
        :param item_class: a :class:`BoundData` subclass to be used for the property value
        """
        def pget(self):
            if not hasattr(self, '__' + data_property):
                setattr(self, '__' + data_property,
                    item_class(
                        path(self._node),
                    )
                )
            return getattr(self, '__' + data_property)

        cls.bind(data_property, pget, None)



from reconfigure.nodes import *
from reconfigure.parsers import BaseParser
import re


class NginxParser (BaseParser):
    """
    A parser for nginx configs
    """

    tokens = [
        (r"[\w_]+\s*?[^\n]*?{", lambda s, t: ('section_start', t)),
        (r"[\w_]+?.+?;", lambda s, t: ('option', t)),
        (r"\s", lambda s, t: 'whitespace'),
        (r"$^", lambda s, t: 'newline'),
        (r"\#.*?\n", lambda s, t: ('comment', t)),
        (r"\}", lambda s, t: 'section_end'),
    ]
    token_comment = '#'
    token_section_end = '}'

    def parse(self, content):
        scanner = re.Scanner(self.tokens, re.DOTALL)
        tokens, remainder = scanner.scan(' '.join(filter(None, content.split(' '))))
        if remainder:
            raise Exception('Invalid tokens: %s. Tokens: %s' % (remainder, tokens))

        node = RootNode()
        node.parameter = None
        node_stack = []
        next_comment = None

        while len(tokens) > 0:
            token = tokens[0]
            tokens = tokens[1:]
            if token in ['whitespace', 'newline']:
                continue
            if token == 'section_end':
                node = node_stack.pop()
            if token[0] == 'comment':
                if not next_comment:
                    next_comment = ''
                else:
                    next_comment += '\n'
                next_comment += token[1].strip('#/*').strip()
            if token[0] == 'option':
                if ' ' in token[1] and not token[1][0] in ['"', "'"]:
                    k, v = token[1].split(None, 1)
                else:
                    v = token[1]
                    k = ''
                prop = PropertyNode(k.strip(), v[:-1].strip())
                prop.comment = next_comment
                next_comment = None
                node.children.append(prop)
            if token[0] == 'section_start':
                line = token[1][:-1].strip().split(None, 1) + [None]
                section = Node(line[0])
                section.parameter = line[1]
                section.comment = next_comment
                next_comment = None
                node_stack += [node]
                node.children.append(section)
                node = section

        return node

    def stringify(self, tree):
        return ''.join(self.stringify_rec(node) for node in tree.children)

    def stringify_rec(self, node):
        if isinstance(node, PropertyNode):
            if node.name:
                s = '%s %s;\n' % (node.name, node.value)
            else:
                s = '%s;\n' % (node.value)
        elif isinstance(node, IncludeNode):
            s = 'include %s;\n' % (node.files)
        else:
            result = '\n%s %s {\n' % (node.name, node.parameter or '')
            for child in node.children:
                result += '\n'.join('\t' + x for x in self.stringify_rec(child).splitlines()) + '\n'
            result += self.token_section_end + '\n'
            s = result
        if node.comment:
            s = ''.join(self.token_comment + ' %s\n' % l for l in node.comment.splitlines()) + s
        return s



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class FSTabData (BoundData):
    pass


class FilesystemData (BoundData):
    fields = ['device', 'mountpoint', 'type', 'options', 'freq', 'passno']

    def template(self):
        return Node('line', children=[
            Node('token', children=[PropertyNode('value', 'none')]),
            Node('token', children=[PropertyNode('value', 'none')]),
            Node('token', children=[PropertyNode('value', 'auto')]),
            Node('token', children=[PropertyNode('value', 'defaults,rw')]),
            Node('token', children=[PropertyNode('value', '0')]),
            Node('token', children=[PropertyNode('value', '0')]),
        ])


FSTabData.bind_collection('filesystems', item_class=FilesystemData)
for i in range(0, len(FilesystemData.fields)):
    path = lambda i: lambda x: x.children[i]
    FilesystemData.bind_property('value', FilesystemData.fields[i], path=path(i))



from reconfigure.builders.base import BaseBuilder


class BoundBuilder (BaseBuilder):
    """
    A builder that uses :class:`reconfigure.items.bound.BoundData` to build stuff

    :param root_class: a ``BoundData`` class that used as processing root
    """

    def __init__(self, root_class):
        self.root_class = root_class

    def build(self, nodetree):
        return self.root_class(nodetree)

    def unbuild(self, tree):
        pass



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class IPTablesData (BoundData):
    pass


class TableData (BoundData):
    def template(self):
        return Node('custom')


class ChainData (BoundData):
    def template(self):
        return Node(
            'CUSTOM',
            PropertyNode('default', '-'),
        )


class RuleData (BoundData):
    def template(self):
        return Node(
            'append',
            Node(
                'option',
                Node('argument', PropertyNode('value', 'ACCEPT')),
                PropertyNode('negative', False),
                PropertyNode('name', 'j'),
            )
        )

    @property
    def summary(self):
        return ' '.join((
            ('! ' if x.negative else '') +
            ('-' if len(x.name) == 1 else '--') + x.name + ' ' +
            ' '.join(a.value for a in x.arguments))
            for x in self.options
        )

    def verify(self):
        protocol_option = None
        for option in self.options:
            if option.name in ['p', 'protocol']:
                self.options.remove(option)
                self.options.insert(0, option)
                protocol_option = option
        for option in self.options:
            if 'port' in option.name:
                if not protocol_option:
                    protocol_option = OptionData.create('protocol')
                    self.options.insert(0, protocol_option)

    def get_option(self, *names):
        for name in names:
            for option in self.options:
                if option.name == name:
                    return option


class OptionData (BoundData):
    templates = {
        'protocol': ['protocol', ['tcp']],
        'match': ['match', ['multiport']],
        'source': ['source', ['127.0.0.1']],
        'mac-source': ['mac-source', ['00:00:00:00:00:00']],
        'destination': ['destination', ['127.0.0.1']],
        'in-interface': ['in-interface', ['lo']],
        'out-interface': ['out-interface', ['lo']],
        'source-port': ['source-port', ['80']],
        'source-ports': ['source-ports', ['80,443']],
        'destination-port': ['destination-port', ['80']],
        'destination-ports': ['destination-ports', ['80,443']],
        'state': ['state', ['NEW']],
        'reject-with': ['reject-with', ['icmp-net-unreachable']],
        'custom': ['name', ['value']],
    }

    @staticmethod
    def create(template_id):
        t = OptionData.templates[template_id]
        return OptionData(Node(
            'option',
            *(
                [Node('argument', PropertyNode('value', x)) for x in t[1]]
                + [PropertyNode('negative', False)]
                + [PropertyNode('name', t[0])]
            )
        ))

    @staticmethod
    def create_destination():
        return OptionData(Node(
            'option',
            Node('argument', PropertyNode('value', 'ACCEPT')),
            PropertyNode('negative', False),
            PropertyNode('name', 'j'),
        ))


class ArgumentData (BoundData):
    pass


IPTablesData.bind_collection('tables', item_class=TableData)
TableData.bind_collection('chains', item_class=ChainData)
TableData.bind_name('name')
ChainData.bind_property('default', 'default')
ChainData.bind_collection('rules', selector=lambda x: x.name == 'append', item_class=RuleData)
ChainData.bind_name('name')
RuleData.bind_collection('options', item_class=OptionData)
RuleData.bind_attribute('comment', 'comment')
OptionData.bind_property('name', 'name')
OptionData.bind_property('negative', 'negative')
OptionData.bind_collection('arguments', selector=lambda x: x.name == 'argument', item_class=ArgumentData)
ArgumentData.bind_property('value', 'value')



from reconfigure.nodes import *
from reconfigure.parsers import BaseParser
from reconfigure.parsers.ssv import SSVParser


class ExportsParser (BaseParser):
    """
    A parser for NFS' /etc/exports
    """

    def __init__(self, *args, **kwargs):
        BaseParser.__init__(self, *args, **kwargs)
        self.inner = SSVParser(continuation='\\')

    def parse(self, content):
        tree = self.inner.parse(content)
        root = RootNode()
        for export in tree:
            export_node = Node(export[0].get('value').value.strip('"'))
            export_node.comment = export.comment
            clients_node = Node('clients')
            export_node.append(clients_node)
            root.append(export_node)

            for client in export[1:]:
                s = client.get('value').value
                name = s.split('(')[0]
                options = ''
                if '(' in s:
                    options = s.split('(', 1)[1].rstrip(')')
                client_node = Node(name)
                client_node.set_property('options', options)
                clients_node.append(client_node)
        return root

    def stringify(self, tree):
        root = RootNode()
        for export in tree:
            export_node = Node('line', comment=export.comment)
            export_node.append(Node('token', PropertyNode('value', '"%s"' % export.name)))
            for client in export['clients']:
                s = client.name
                if client['options'].value:
                    s += '(%s)' % client['options'].value
                export_node.append(Node('token', PropertyNode('value', s)))
            root.append(export_node)
        return self.inner.stringify(root)



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class GroupsData (BoundData):
    pass


class GroupData (BoundData):
    fields = ['name', 'password', 'gid', 'users']

    def template(self):
        return Node(
            'line',
            *[
                Node('token', children=[
                    PropertyNode('value', '')
                ])
                for x in GroupData.fields
            ]
        )


GroupsData.bind_collection('groups', item_class=GroupData)
for i in range(0, len(GroupData.fields)):
    path = lambda i: lambda x: x.children[i]
    GroupData.bind_property('value', GroupData.fields[i], path=path(i))



from reconfigure.nodes import *
from reconfigure.parsers import BaseParser


class SquidParser (BaseParser):
    """
    A parser for Squid configs
    """

    def parse(self, content):
        lines = filter(None, [x.strip() for x in content.splitlines()])
        root = RootNode()
        last_comment = None
        for line in lines:
            line = line.strip()
            if line.startswith('#'):
                c = line.strip('#').strip()
                if last_comment:
                    last_comment += '\n' + c
                else:
                    last_comment = c
                continue
            if len(line) == 0:
                continue
            tokens = line.split()
            node = Node('line', Node('arguments'))
            if last_comment:
                node.comment = last_comment
                last_comment = None

            index = 0
            for token in tokens:
                if token.startswith('#'):
                    node.comment = ' '.join(tokens[tokens.index(token):])[1:].strip()
                    break
                if index == 0:
                    node.set_property('name', token)
                else:
                    node.get('arguments').set_property(str(index), token)
                index += 1
            root.append(node)
        return root

    def stringify(self, tree):
        r = ''
        for node in tree.children:
            if node.comment and '\n' in node.comment:
                r += ''.join('%s %s\n' % ('#', x) for x in node.comment.splitlines())
            r += node.get('name').value + ' ' + ' '.join(x.value for x in node.get('arguments').children)
            if node.comment and not '\n' in node.comment:
                r += ' # %s' % node.comment
            r += '\n'
        return r



from reconfigure.items.bound import BoundData
from reconfigure.nodes import Node, PropertyNode
import unittest


class BoundDataTest (unittest.TestCase):

    def test_bind_property(self):
        class TestBoundData (BoundData):
            pass
        TestBoundData.bind_property('prop', 'dataprop', getter=lambda x: 'd' + x, setter=lambda x: x[1:])

        n = Node('name', children=[
                PropertyNode('prop', 'value')
            ])

        d = TestBoundData(n)

        self.assertEqual(d.dataprop, 'dvalue')
        d.dataprop = 'dnew'
        self.assertEqual(d.dataprop, 'dnew')
        self.assertEqual(n.get('prop').value, 'new')

    def test_bind_collection(self):
        class TestBoundData (BoundData):
            pass

        class TestChildData (BoundData):
            def template(self):
                return Node('', children=[PropertyNode('value', None)])

        TestBoundData.bind_collection('items', item_class=TestChildData, selector=lambda x: x.name != 'test')
        TestChildData.bind_property('value', 'value')
        n = Node('name', children=[
                Node('1', children=[PropertyNode('value', 1)]),
                Node('2', children=[PropertyNode('value', 2)]),
                Node('test', children=[PropertyNode('value', 3)]),
                Node('3', children=[PropertyNode('value', 3)]),
            ])

        d = TestBoundData(n)
        self.assertEqual(d.items[0].value, 1)
        self.assertEqual(len(d.items), 3)
        c = TestChildData()
        c.value = 4
        d.items.append(c)
        self.assertEqual(len(d.items), 4)
        self.assertEqual(d.items[-1].value, 4)


if __name__ == '__main__':
    unittest.main()



import json

from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData, BoundDictionary


class AjentiData (BoundData):
    pass


class HttpData (BoundData):
    pass


class SSLData (BoundData):
    pass


class UserData (BoundData):
    def template(self):
        return Node(
            'unnamed',
            PropertyNode('configs', {}),
            PropertyNode('password', ''),
            PropertyNode('permissions', []),
        )


class ConfigData (BoundData):
    def template(self):
        return PropertyNode('', '{}')


AjentiData.bind_property('authentication', 'authentication')
AjentiData.bind_property('language', 'language')
AjentiData.bind_property('installation_id', 'installation_id')
AjentiData.bind_property('enable_feedback', 'enable_feedback')
AjentiData.bind_child('http_binding', lambda x: x.get('bind'), item_class=HttpData)
AjentiData.bind_child('ssl', lambda x: x.get('ssl'), item_class=SSLData)
AjentiData.bind_collection('users', path=lambda x: x.get('users'), item_class=UserData, collection_class=BoundDictionary, key=lambda x: x.name)


HttpData.bind_property('host', 'host')
HttpData.bind_property('port', 'port')

SSLData.bind_property('certificate_path', 'certificate_path')
SSLData.bind_property('enable', 'enable')

ConfigData.bind_name('name')

UserData.bind_name('name')
UserData.bind_property('email', 'email')
UserData.bind_property('password', 'password')
UserData.bind_property('permissions', 'permissions')
UserData.bind_collection('configs', lambda x: x.get('configs'), item_class=ConfigData, collection_class=BoundDictionary, key=lambda x: x.name)

ConfigData.bind_attribute('value', 'data', getter=json.loads, setter=json.dumps)



"""
Builders are used to convert Node Tree to Data Tree
"""

from reconfigure.builders.base import BaseBuilder
from reconfigure.builders.bound import BoundBuilder


__all__ = [
    'BaseBuilder',
    'BoundBuilder',
]



import chardet
import sys


class Reconfig (object):
    """
    Basic config class. Derivatives normally only need to override the constructor.

    Config data is loaded either from ``path`` or from ``content``

    :param parser: overrides the Parser instance
    :param includer: overrides the Includer instance
    :param builder: overrides the Builder instance
    :param path: config file path. Not compatible with ``content``
    :param content: config file content. Not compatible with ``path``
    """

    def __init__(self, parser=None, includer=None, builder=None, path=None, content=None):
        self.parser = parser
        self.builder = builder
        self.includer = includer
        if self.includer is not None:
            if not self.includer.parser:
                self.includer.parser = self.parser
        if path:
            self.origin = path
            self.content = None
        else:
            self.origin = None
            self.content = content

    def load(self):
        """
        Loads the config data, parses and builds it. Sets ``tree`` attribute to point to Data tree.
        """
        if self.origin:
            self.content = open(self.origin, 'r').read()

        self.encoding = 'utf8'
        if (sys.version_info[0] >= 3 and isinstance(self.content, bytes)) or \
                (sys.version_info[0] == 2 and isinstance(self.content, str)):
            try:
                self.content = self.content.decode('utf8')
            except (UnicodeDecodeError, AttributeError):
                self.encoding = chardet.detect(self.content)['encoding']
                self.content = self.content.decode(self.encoding)

        self.nodetree = self.parser.parse(self.content)
        if self.includer is not None:
            self.nodetree = self.includer.compose(self.origin, self.nodetree)
        if self.builder is not None:
            self.tree = self.builder.build(self.nodetree)
        return self

    def save(self):
        """
        Unbuilds, stringifies and saves the config. If the config was loaded from string, returns ``{ origin: data }`` dict
        """
        tree = self.tree
        if self.builder is not None:
            nodetree = self.builder.unbuild(tree) or self.nodetree
        if self.includer is not None:
            nodetree = self.includer.decompose(nodetree)
        else:
            nodetree = {self.origin: nodetree}

        result = {}
        for k in nodetree:
            v = self.parser.stringify(nodetree[k])
            if self.encoding != 'utf8':
                v = v.encode(self.encoding)
            result[k or self.origin] = v

        if self.origin is not None:
            for k in result:
                open(k, 'w').write(result[k])
        return result



#!/usr/bin/env python

from distutils.core import setup
from setuptools import find_packages

from reconfigure import __version__

setup(
    name='reconfigure',
    version=__version__,
    install_requires=[
        'chardet'
    ],
    description='An ORM for config files',
    license='LGPLv3',
    author='Eugeny Pankov',
    author_email='e@ajenti.org',
    url='http://ajenti.org/',
    classifiers=[
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    ],
    packages=find_packages(exclude=['*test*']),
)



from reconfigure.configs.base import Reconfig
from reconfigure.parsers import IPTablesParser
from reconfigure.builders import BoundBuilder
from reconfigure.items.iptables import IPTablesData


class IPTablesConfig (Reconfig):
    """
    ``iptables-save`` and ``iptables-restore``
    """
    def __init__(self, **kwargs):
        k = {
            'parser': IPTablesParser(),
            'builder': BoundBuilder(IPTablesData),
        }
        k.update(kwargs)
        Reconfig.__init__(self, **k)



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class ExportsData (BoundData):
    pass


class ExportData (BoundData):
    def template(self):
        return Node(
            '/',
            Node('clients')
        )


class ClientData (BoundData):
    def template(self):
        return Node(
            'localhost',
            PropertyNode('options', '')
        )


ExportsData.bind_collection('exports', item_class=ExportData)
ExportData.bind_name('name')
ExportData.bind_attribute('comment', 'comment', default='')
ExportData.bind_collection('clients', path=lambda x: x['clients'], item_class=ClientData)
ClientData.bind_name('name')
ClientData.bind_property('options', 'options')



from reconfigure.nodes import Node, PropertyNode
from reconfigure.items.bound import BoundData


class SquidData (BoundData):
    pass


class ACLData (BoundData):
    def template(self, name=None, *args):
        children = [PropertyNode('1', name)]
        index = 2
        for arg in args:
            children += [PropertyNode(str(index), arg)]
            index += 1
        return Node(
            'line',
            PropertyNode('name', 'acl'),
            Node(
                'arguments',
                *children
            )
        )

    def describe(self):
        return ' '.join(x.value for x in self.options)


class HTTPAccessData (BoundData):
    def template(self):
        return Node(
            'line',
            PropertyNode('name', 'http_access'),
            Node('arguments', PropertyNode('1', ''))
        )


class HTTPPortData (BoundData):
    def template(self):
        return Node(
            'line',
            PropertyNode('name', 'http_port'),
            Node('arguments', PropertyNode('1', '3128'))
        )


class HTTPSPortData (BoundData):
    def template(self):
        return Node(
            'line',
            PropertyNode('name', 'https_port'),
            Node('arguments', PropertyNode('1', '3128'))
        )


class ArgumentData (BoundData):
    def template(self):
        return PropertyNode('value', 'none')


def __bind_by_name(cls, prop, name, itemcls):
    cls.bind_collection(
        prop,
        selector=lambda x: x.get('name').value == name,
        item_class=itemcls
    )

__bind_by_name(SquidData, 'acl', 'acl', ACLData)
__bind_by_name(SquidData, 'http_access', 'http_access', HTTPAccessData)
__bind_by_name(SquidData, 'http_port', 'http_port', HTTPPortData)
__bind_by_name(SquidData, 'https_port', 'https_port', HTTPSPortData)


def __bind_first_arg(cls, prop):
    cls.bind_attribute('value', prop, path=lambda x: x.get('arguments').children[0])


def __bind_other_args(cls, prop, itemcls):
    cls.bind_collection(
        prop, path=lambda x: x.get('arguments'),
        selector=lambda x: x.parent.children.index(x) > 0, item_class=itemcls
    )

__bind_first_arg(ACLData, 'name')
__bind_other_args(ACLData, 'options', ArgumentData)

__bind_first_arg(HTTPAccessData, 'mode')
__bind_other_args(HTTPAccessData, 'options', ArgumentData)

__bind_first_arg(HTTPPortData, 'port')
__bind_other_args(HTTPPortData, 'options', ArgumentData)
__bind_first_arg(HTTPSPortData, 'port')
__bind_other_args(HTTPSPortData, 'options', ArgumentData)

ArgumentData.bind_attribute('value', 'value')



from reconfigure.includers.auto import AutoIncluder
from reconfigure.nodes import PropertyNode


class NginxIncluder (AutoIncluder):
    def is_include(self, node):
        if isinstance(node, PropertyNode) and node.name == 'include':
            return node.value

    def remove_include(self, node):
        return PropertyNode('include', node.files)
